/*--------------------------------------------------------------------------
**
**  Copyright (c) 2003, Ken Hunter 
**
**  Name: ecs_win32.c
**
**  Description:
**      Simulate ECS as Windows virtual memory, with a named file as
**      the backing store, not the system paging file.  This will allow
**		the contents of ECS to be persistent across deadstarts.
**
**		While I have not studied the UNIX interface for doing similar
**		things, I'd guess that it's not the same as for Windows.  Most
**		likely, this code will only work on the Win32 platform.
**
**--------------------------------------------------------------------------
*/

/*
**  -------------
**  Include Files
**  -------------
*/
#include <stdio.h>
#include <windows.h>
#include "resource.h"
#include "const.h"
#include "types.h"
#include "proto.h"

/*
**  -----------------
**  Private Constants
**  -----------------
*/

#define ECSFileName			"SYSTEM_ECS"
#define ECSFileAccess		(GENERIC_READ | GENERIC_WRITE)
#define ECSFileSharing		(FILE_SHARE_READ | FILE_SHARE_WRITE)
#define ECSFileSecurity		NULL
#define ECSFileCreation		OPEN_ALWAYS
#define ECSFileFlags		(FILE_ATTRIBUTE_NORMAL | FILE_FLAG_RANDOM_ACCESS)
#define ECSFileTemplate		NULL
#define ECSMapSecurity		NULL
#define ECSMapProtect		(PAGE_READWRITE | SEC_COMMIT)
#define ECSMapSizeHigh		0
#define ECSMapName			NULL
#define ECSViewAccess		FILE_MAP_ALL_ACCESS
#define ECSViewOffsetHigh	0
#define ECSViewOffsetLow	0

/*
**  -----------------------
**  Private Macro Functions
**  -----------------------
*/

/*
**  -----------------------------------------
**  Private Typedef and Structure Definitions
**  -----------------------------------------
*/

/*
**  ---------------------------
**  Private Function Prototypes
**  ---------------------------
*/

/*
**  ----------------
**  Public Variables
**  ----------------
*/

/*
**  -----------------
**  Private Variables
**  -----------------
*/

static HANDLE ecsFile;
static HANDLE ecsMap;

/*--------------------------------------------------------------------------
**  Purpose:        Establish ECS in Windows virtual memory.
**
**  Parameters:     Name        Description.
**                  ecsSize     size of ECS in CYBER words.
**
**  Returns:        Nothing.
**
**	On exit, the global variable ecsMem is set to point to the first
**	CYBER word of ECS, if ECS is successfully established.  If ECS is
**	not successfully established, ecsMem is set to NULL, and all Windows
**	objects relating to ECS are released.
**
**------------------------------------------------------------------------*/
void openECS(u32 ecsSize)
{
	DWORD  ecsByteSize;

	ecsMem = NULL;

	ecsFile = CreateFile(ECSFileName, ECSFileAccess, ECSFileSharing,
		ECSFileSecurity, ECSFileCreation, ECSFileFlags, ECSFileTemplate);
	if (INVALID_HANDLE_VALUE == ecsFile)
		return;
	ecsByteSize = ecsSize * sizeof(CpWord);
	ecsMap = CreateFileMapping(ecsFile, ECSMapSecurity, ECSMapProtect,
		ECSMapSizeHigh, ecsByteSize, ECSMapName);
	if (NULL == ecsMap)
	{
		CloseHandle(ecsFile);
		return;
	}
	ecsMem = (CpWord *)MapViewOfFile(ecsMap, ECSViewAccess,
		ECSViewOffsetHigh, ECSViewOffsetLow, ecsByteSize);
	if (NULL == ecsMem)
	{
		CloseHandle(ecsMap);
		CloseHandle(ecsFile);
	}
}
/*--------------------------------------------------------------------------
**  Purpose:        Release ECS virtual memory, and close the backing file..
**
**  Parameters:     None.
**
**  Returns:        Nothing.
**
**	On exit, the backing file for ECS is closed, and all Windows objects
**	relating to ECS are released.
**
**	If winCloseECS is not called, as might happen if the entire DtCyber
**	process is aborted, it is possible that some or all of the contents
**	of ECS will not be paged out to disk.
**
**------------------------------------------------------------------------*/
void closeECS(void)
{
	if (NULL != ecsMem)
	{
		UnmapViewOfFile((LPCVOID)ecsMem);
		CloseHandle(ecsMap);
		CloseHandle(ecsFile);
	}
}
/*---------------------------  End Of File  ------------------------------*/
