/*--------------------------------------------------------------------------
**
**  Copyright (c) 2003-2005, Tom Hunter (see license.txt)
**
**  Name: maintenance_channel.c
**
**  Description:
**      Perform emulation of maintenance channel.
**
**--------------------------------------------------------------------------
*/

#define DEBUG 0

/*
**  -------------
**  Include Files
**  -------------
*/
#include <stdio.h>
#include <stdlib.h>
#ifndef WIN32
#include <unistd.h>
#endif
#include <string.h>
#include "const.h"
#include "types.h"
#include "proto.h"

/*
**  -----------------
**  Private Constants
**  -----------------
*/

/*
**  Connect codes.
*/
#define FcConnIou               0x000
#define FcConnMemory            0x100
#define FcConnCpu               0x200
#define FcConnIdleLow           0x800
#define FcConnIdleHigh          0xF00

#define FcConnMask              0xF00

/*
**  Function codes.
*/
#define FcOpHalt                0x000
#define FcOpStart               0x010
#define FcOpClearLed            0x030
#define FcOpRead                0x040
#define FcOpWrite               0x050
#define FcOpMasterClear         0x060
#define FcOpClearErrors         0x070
#define FcOpEchoData            0x080
#define FcOpRequSummaryStatus   0x0C0

#define FcOpMask                0x0F0

/*
**  Type bits.
*/
#define FcTypeMask              0x00F

/*
**  Register addresses.
*/
#define RegAddrSummaryStatus    0x00
#define RegAddrElementId        0x10
#define RegAddrOptionsInstalled 0x12
#define RegAddrFaultStatus      0x18
#define RegAddrOsBounds         0x21
#define RegAddrEnvControl       0x30
#define RegAddrStatus           0x40
#define RegAddrFaultStatus1     0x80
#define RegAddrFaultStatus2     0x81
#define RegAddrTestMode         0xA0

/*
**  -----------------------
**  Private Macro Functions
**  -----------------------
*/

/*
**  -----------------------------------------
**  Private Typedef and Structure Definitions
**  -----------------------------------------
*/

/*
**  ---------------------------
**  Private Function Prototypes
**  ---------------------------
*/
static FcStatus mchFunc(PpWord funcCode);
static void mchIo(void);
static void mchActivate(void);
static void mchDisconnect(void);
static char *mchConn2String(PpWord connCode);
static char *mchOp2String(PpWord opCode);

/*
**  ----------------
**  Public Variables
**  ----------------
*/

/*
**  -----------------
**  Private Variables
**  -----------------
*/
static u8 mchLocation;
static bool mchAddressReady;
static CpWord dataIou[256];
static CpWord dataMem[256];
static CpWord dataCpu[256];

#if DEBUG
static FILE *mchLog = NULL;
#endif

/*
**--------------------------------------------------------------------------
**
**  Public Functions
**
**--------------------------------------------------------------------------
*/
/*--------------------------------------------------------------------------
**  Purpose:        Initialise maintenance channel.
**
**  Parameters:     Name        Description.
**                  eqNo        equipment number
**                  unitNo      unit number
**                  channelNo   channel number the device is attached to
**                  deviceName  optional device file name
**
**  Returns:        Nothing.
**
**------------------------------------------------------------------------*/
void mchInit(u8 eqNo, u8 unitNo, u8 channelNo, char *deviceName)
    {
    DevSlot *dp;

    (void)eqNo;
    (void)deviceName;

#if DEBUG
    if (mchLog == NULL)
        {
        mchLog = fopen("mchlog.txt", "wt");
        }
#endif

    dp = channelAttach(channelNo, eqNo, DtMch);
    dp->activate = mchActivate;
    dp->disconnect = mchDisconnect;
    dp->func = mchFunc;
    dp->io = mchIo;

#if 0
    /*
    **  Initialize registers (representing a 840A with 2 barrels with
    **  PP0 - PP11, channels 0 - 17, TPM and CC545, 16 MB of memory).
    */                             
    //                                   0011223244556677
    dataIou[RegAddrElementId       ] = 0x0000000002201234;
    dataMem[RegAddrElementId       ] = 0x0000000001311234;
    dataCpu[RegAddrElementId       ] = 0x0000000000341234;

    //                                   0011223244556677
    dataIou[RegAddrOptionsInstalled] = 0x000003FFFF00000F;
    dataMem[RegAddrOptionsInstalled] = 0x0010000000000000;
    dataCpu[RegAddrOptionsInstalled] = 0x0000000000000000;
#else
    /*
    **  Initialize registers (representing a 860 with 2 barrels with
    **  PP0 - PP11, channels 0 - 17, TPM and CC545, 16 MB of memory).
    */                             
    //                                   0011223244556677
    dataIou[RegAddrElementId       ] = 0x0000000002201234;
    dataMem[RegAddrElementId       ] = 0x0000000001311234;
    dataCpu[RegAddrElementId       ] = 0x0000000000321234;

    //                                   0011223244556677
    dataIou[RegAddrOptionsInstalled] = 0x000003FFAF000007;
    dataMem[RegAddrOptionsInstalled] = 0x0010000000000000;
    dataCpu[RegAddrOptionsInstalled] = 0x0000000000000000;
#endif

    /*
    **  Print a friendly message.
    */
    printf("Maintenance Channel initialised on channel %o\n", channelNo);
    }

/*--------------------------------------------------------------------------
**  Purpose:        Execute function code on maintenance channel.
**
**  Parameters:     Name        Description.
**                  funcCode    function code
**
**  Returns:        FcStatus
**
**------------------------------------------------------------------------*/
static FcStatus mchFunc(PpWord funcCode)
    {
    PpWord connCode;
    PpWord opCode;
    PpWord typeCode;

    connCode = funcCode & FcConnMask;
    opCode   = funcCode & FcOpMask;
    typeCode = funcCode & FcTypeMask;

#if DEBUG
    fprintf(mchLog, "\n%06d PP:%02o CH:%02o f:0x%03X C:%-25s O:%-25s  >   ",
        traceSequenceNo,
        activePpu->id,
        activeDevice->channel->id,
        funcCode,
        mchConn2String(connCode),
        mchOp2String(opCode));
#endif

    /*
    **  Process connect codes. Note that 0x800 - 0xF00
    **  causes the MCH to be deselected (we do it for all
    **  connect codes we don't know about).
    */
    switch (connCode)
        {
    case FcConnIou:
    case FcConnMemory:
    case FcConnCpu:
        break;

    default:
        return(FcProcessed);
        }

    /*
    **  Process operation codes.
    */
    switch (opCode)
        {
    default:
#if DEBUG
        fprintf(mchLog, " Operation not implemented & declined!");
#endif
        return(FcDeclined);

    case FcOpHalt:
    case FcOpStart:
    case FcOpClearLed:
    case FcOpMasterClear:
    case FcOpClearErrors:
        /*
        **  Do nothing.
        */
        return(FcProcessed);

    case FcOpRead:
    case FcOpWrite:
    case FcOpEchoData:
        mchLocation = 0;
        mchAddressReady = FALSE;
        activeDevice->recordLength = 2;
        break;
        
    case FcOpRequSummaryStatus:
        activeDevice->recordLength = 1;
        break;
        }

    activeDevice->fcode = funcCode;
    return(FcAccepted);
    }

/*--------------------------------------------------------------------------
**  Purpose:        Perform I/O on maintenance channel.
**
**  Parameters:     Name        Description.
**
**  Returns:        Nothing.
**
**------------------------------------------------------------------------*/
static void mchIo(void)
    {
    PpWord connCode;
    PpWord opCode;
    PpWord typeCode;
    CpWord *dp;
    int shiftCount;

    connCode = activeDevice->fcode & FcConnMask;
    opCode   = activeDevice->fcode & FcOpMask;
    typeCode = activeDevice->fcode & FcTypeMask;

    switch (opCode)
        {
    default:
        break;

    case FcOpRead:
        if (!mchAddressReady)
            {
            if (activeChannel->full)
                {
#if DEBUG
                fprintf(mchLog, " %02X", activeChannel->data);
#endif
                activeChannel->full = FALSE;
                activeDevice->recordLength -= 1;

                if (activeDevice->recordLength == 0)
                    {
                    /*
                    **  Ignore the MSB of the address.
                    */
                    mchLocation = (activeChannel->data & Mask8);
                    }
                }
            }
        else
            {
            if (!activeChannel->full)
                {
                switch (connCode)
                    {
                default:
                case FcConnIou:
                    dp = dataIou;
                    break;

                case FcConnMemory:
                    dp = dataMem;
                    break;

                case FcConnCpu:
                    dp = dataCpu;
                    break;
                    }

                switch (mchLocation)
                    {
                case RegAddrElementId:
                case RegAddrOptionsInstalled:
                    typeCode = 0;
                    break;

                default:
                    break;
                    }

                if (activeDevice->recordLength > 0)
                    {
                    activeDevice->recordLength -= 1;
                    shiftCount = ((activeDevice->recordLength + typeCode) % 8) * 8;

                    activeChannel->data = (PpWord)((dp[mchLocation] >> shiftCount) & Mask8);
                    }
                else
                    {
                    activeChannel->data = 0;
                    }
                activeChannel->full = TRUE;
#if DEBUG
                fprintf(mchLog, " %02X", activeChannel->data);
#endif
                }
            }

        break;

    case FcOpWrite:
        if (!mchAddressReady)
            {
            if (activeChannel->full)
                {
#if DEBUG
                fprintf(mchLog, " %02X", activeChannel->data);
#endif
                activeChannel->full = FALSE;
                activeDevice->recordLength -= 1;

                if (activeDevice->recordLength == 0)
                    {
                    /*
                    **  Ignore the MSB of the address.
                    */
                    mchLocation = (activeChannel->data & Mask8);
                    }
                }
            }
        else
            {
            if (activeChannel->full)
                {
                switch (connCode)
                    {
                default:
                case FcConnIou:
                    dp = dataIou;
                    break;

                case FcConnMemory:
                    dp = dataMem;
                    break;

                case FcConnCpu:
                    dp = dataCpu;
                    break;
                    }

                switch (mchLocation)
                    {
                case RegAddrElementId:
                case RegAddrOptionsInstalled:
                    typeCode = 0;
                    break;

                default:
                    break;
                    }

                if (activeDevice->recordLength > 0)
                    {
                    activeDevice->recordLength -= 1;
                    shiftCount = ((activeDevice->recordLength + typeCode) % 8) * 8;

                    dp[mchLocation] &= ~((CpWord)Mask8 << shiftCount);
                    dp[mchLocation] |= ((CpWord)activeChannel->data & Mask8) << shiftCount;
                    }
#if DEBUG
                fprintf(mchLog, " %02X", activeChannel->data);
#endif
                activeChannel->full = FALSE;
                }
            }

        break;

    case FcOpEchoData:
        if (!mchAddressReady)
            {
            if (activeChannel->full)
                {
#if DEBUG
                fprintf(mchLog, " %02X", activeChannel->data);
#endif
                activeChannel->full = FALSE;
                activeDevice->recordLength -= 1;

                if (activeDevice->recordLength == 0)
                    {
                    /*
                    **  Ignore the MSB of the address.
                    */
                    mchLocation = (activeChannel->data & Mask8);
                    }
                }
            }
        else
            {
            if (!activeChannel->full)
                {
                activeChannel->data = (PpWord)mchLocation;
                activeChannel->full = TRUE;
#if DEBUG
                fprintf(mchLog, " %02X", activeChannel->data);
#endif
                }
            }

        break;

    case FcOpRequSummaryStatus:
        if (!activeChannel->full)
            {
            activeChannel->data = 0;
            activeChannel->full = TRUE;
            }

        break;
        }
    }

/*--------------------------------------------------------------------------
**  Purpose:        Handle channel activation.
**
**  Parameters:     Name        Description.
**
**  Returns:        Nothing.
**
**------------------------------------------------------------------------*/
static void mchActivate(void)
    {
    }

/*--------------------------------------------------------------------------
**  Purpose:        Handle disconnecting of channel.
**
**  Parameters:     Name        Description.
**
**  Returns:        Nothing.
**
**------------------------------------------------------------------------*/
static void mchDisconnect(void)
    {
    switch (activeDevice->fcode & FcOpMask)
        {
    case FcOpRead:
    case FcOpWrite:
        if (!mchAddressReady)
            {
            mchAddressReady = TRUE;
            activeDevice->recordLength = 8;
            }

        break;
        }
    }

/*--------------------------------------------------------------------------
**  Purpose:        Convert connect code to string.
**
**  Parameters:     Name        Description.
**                  connCode    operation code
**
**  Returns:        String equivalent of connect code.
**
**------------------------------------------------------------------------*/
static char *mchConn2String(PpWord connCode)
    {
    static char buf[30];
#if DEBUG
    switch(connCode)
        {
    case FcConnIou         : return "IOU";
    case FcConnMemory      : return "Memory";
    case FcConnCpu         : return "Processor";
        }
#endif
    sprintf(buf, "UNKNOWN: %03X", connCode);
    return(buf);
    }

/*--------------------------------------------------------------------------
**  Purpose:        Convert operation code to string.
**
**  Parameters:     Name        Description.
**                  opCode      operation code
**
**  Returns:        String equivalent of operation code.
**
**------------------------------------------------------------------------*/
static char *mchOp2String(PpWord opCode)
    {
    static char buf[30];
#if DEBUG
    switch(opCode)
        {
    case FcOpHalt              : return "Halt";
    case FcOpStart             : return "Start";
    case FcOpClearLed          : return "ClearLed";
    case FcOpRead              : return "Read";
    case FcOpWrite             : return "Write";
    case FcOpMasterClear       : return "MasterClear";
    case FcOpClearErrors       : return "ClearErrors";
    case FcOpEchoData          : return "EchoData";
    case FcOpRequSummaryStatus : return "RequSummaryStatus";
        }
#endif
    sprintf(buf, "UNKNOWN: %03X", opCode);
    return(buf);
    }

/*---------------------------  End Of File  ------------------------------*/
